from syft.frameworks.torch.tensors.interpreters.paillier import PaillierTensor
from syft.serde.serde import deserialize
from phe.paillier import EncryptedNumber
from phe.paillier import PaillierPublicKey
from json import JSONEncoder
from numpy import ndarray
from numpy import array
from numpy import random
from web3 import Web3

from syft import TorchHook
import torch
hook = TorchHook(torch)

# User manipulation functions
def retrieve_user_info(user_path = 'info/user_info.txt'):
    user_file = open(user_path, 'r')
    address = user_file.readline().lstrip('address:').rstrip()
    private_key = user_file.readline().lstrip('private_key:').rstrip()
    endpoint_url = user_file.readline().lstrip('endpoint_url:').rstrip()
    user_file.close()
    return address, private_key, endpoint_url

# Paillier tensor manipulation functions
class MultiDimensionalArrayEncoder(JSONEncoder):
    def encode(self, obj):
        def hint_tuples(item):
            if isinstance(item, tuple):
                return {'__tuple__': True, 'items': item}
            if isinstance(item, list):
                return [hint_tuples(e) for e in item]
            if isinstance(item, dict):
                return {key: hint_tuples(value) for key, value in item.items()}
            else:
                return item
        return super(MultiDimensionalArrayEncoder, self).encode(hint_tuples(obj))
def hinted_tuple_hook(obj):
    if '__tuple__' in obj:
        return tuple(obj['items'])
    else:
        return obj
def serialize_paillier(element):
  # Case 1: tensor recursion
  if isinstance(element, torch.Tensor):
    paillier = element.child
    if isinstance(paillier, PaillierTensor):
      child = [serialize_paillier(subchild) for subchild in paillier.child]
      return {'n': paillier.pubkey.n, 'values': child} # in PaillierPublicKey g = n + 1
    else:
      raise TypeError(type(paillier))

  # Case 2: ndarray recursion
  elif isinstance(element, ndarray):
    return [serialize_paillier(subelement) for subelement in element]

  # Case 3: EncryptedNumber serialization
  elif isinstance(element, EncryptedNumber):
    return (str(element.ciphertext()), str(element.exponent))

  # Case 4: Unknown type
  else:
    raise TypeError(type(element))
def deserialize_paillier(struct, pub=None):
  # Case 1: dict recursion
  if isinstance(struct, dict):
    pub = PaillierPublicKey(n=int(struct['n']))
    child = [deserialize_paillier(substruct, pub) for substruct in struct['values']]
    # Building Paillier Tensor
    tensor = PaillierTensor()
    tensor.child = array(child)
    tensor.pubkey = pub
    return tensor.wrap()

  # Case 2: list recursion
  elif isinstance(struct, list):
    return [deserialize_paillier(substruct, pub) for substruct in struct]

  # Case 3: Tuple deserialization
  elif isinstance(struct, tuple):
    return EncryptedNumber(pub, int(struct[0]), int(struct[1]))

  # Case 4: Unknown type
  else:
    raise TypeError(type(struct))
def to_paillier(element, public_key):
    if isinstance(element, torch.Tensor) and isinstance(element.child, PaillierTensor):
        element.child.pubkey = public_key
        child = element.child.child
        if isinstance(child, ndarray):
          return element
        elif isinstance(child, EncryptedNumber):
          element.child.child = array([child])
          return element
        else:
          raise Exception("The tensor does not have an EncryptedNumber or a np.ndarray as child")
    elif isinstance(element, PaillierTensor):
        element.pubkey = public_key
        return element.wrap()
    elif isinstance(element, ndarray):
        tensor = PaillierTensor()
        tensor.child = element
        tensor.pubkey = public_key
        return tensor.wrap()
    elif isinstance(element, list):
        tensor = PaillierTensor()
        tensor.child = array(element)
        tensor.pubkey = public_key
        return tensor.wrap()
    elif isinstance(element, EncryptedNumber):
        tensor = PaillierTensor()
        tensor.child = array([element])
        tensor.pubkey = public_key
        return tensor.wrap()
    else:
        raise TypeError(type(element))

# Configuration functions
def set_web3(infura_url, my_address):
    web3 = Web3(Web3.HTTPProvider(infura_url))
    web3.eth.defaultAccount = my_address
    return web3



from apscheduler.schedulers.background import BackgroundScheduler
from atexit import register
from os import environ
from psycopg2 import connect
from urllib.parse import urlparse

from numpy import ndarray
from numpy import array

from hesystem.essentials import retrieve_user_info
from hesystem.essentials import MultiDimensionalArrayEncoder
from hesystem.essentials import hinted_tuple_hook
from hesystem.essentials import serialize_paillier
from hesystem.essentials import deserialize_paillier
from hesystem.essentials import to_paillier
from hesystem.essentials import set_web3

# Contract information container
class ContractData():
    def __init__(self, value, expiration_time, meta_link, test_link, test_result_link, data_link, result_link):
        self.value = value
        self.guarantee = int(value*0.1)
        self.expiration_time = expiration_time
        self.meta_link = meta_link
        self.test_link = test_link
        self.test_result_link = test_result_link
        self.data_link = data_link
        self.result_link = result_link

        abi_file = open('info/abi.txt','r')
        bytecode_file = open('info/bytecode.txt','r')
        self.abi = abi_file.read().rstrip()
        self.bytecode = bytecode_file.read().rstrip()
    def deploy_contract(self, web3, my_address, private_key):
        Contract = web3.eth.contract(abi=self.abi, bytecode=self.bytecode)
        construct_tx = Contract.constructor(
            self.value,
            self.guarantee,
            self.expiration_time,
            self.meta_link,
            self.test_link,
            self.test_result_link,
            self.data_link,
            self.result_link,
        ).buildTransaction({
            'nonce': web3.eth.getTransactionCount(my_address),
            'gas': 1728712,
            'gasPrice': web3.toWei('21','gwei')
        })
        signed_tx = web3.eth.account.signTransaction(construct_tx, private_key=private_key)
        tx_hash = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
        tx_receipt = web3.eth.waitForTransactionReceipt(tx_hash)
        contract_address = tx_receipt.contractAddress
        address = web3.toChecksumAddress(contract_address)
        contract_deployed = web3.eth.contract(address=address,abi=self.abi)
        self.address = contract_address
        return contract_deployed
    def use_contract(self, web3, contract_address):
        self.address = contract_address
        address = web3.toChecksumAddress(contract_address)
        contract_deployed = web3.eth.contract(address=address,abi=self.abi)
        return contract_deployed
def retrieve_contract_info(contract_path='info/contract_info.txt'):
    contract_file = open(contract_path, 'r')
    meta_link = contract_file.readline().lstrip('meta_link:').rstrip()
    test_link = contract_file.readline().lstrip('test_link:').rstrip()
    test_result_link = contract_file.readline().lstrip('test_result_link:').rstrip()
    data_link = contract_file.readline().lstrip('data_link:').rstrip()
    result_link = contract_file.readline().lstrip('result_link:').rstrip()
    num_rows = eval(contract_file.readline().lstrip('num_rows:').rstrip())
    value = eval(contract_file.readline().lstrip('value:').rstrip())
    expiration_time = eval(contract_file.readline().lstrip('expiration_time:').rstrip())
    condition = eval(contract_file.readline().lstrip('condition:').rstrip())
    contract_file.close()
    return meta_link, test_link, test_result_link, data_link, result_link, num_rows, value, expiration_time, condition

# Paillier comparison functions
def greather_than(tensor):
    if isinstance(tensor, ndarray):
        result = tensor
        for i,e in enumerate(tensor):
            result[i] = greather_than(tensor[i])
        return result
    else:
        if tensor > 0:
            return True
        else:
            return False
def less_than(tensor):
    if isinstance(tensor, ndarray):
        result = tensor
        for i,e in enumerate(tensor):
            result[i] = less_than(tensor[i])
        return result
    else:
        if tensor < 0:
            return True
        else:
            return False
def equal_to(tensor):
    if isinstance(tensor, ndarray):
        result = tensor
        for i,e in enumerate(tensor):
            result[i] = equal_to(tensor[i])
        return result
    else:
        if tensor == 0:
            return True
        else:
            return False
def equal_great(tensor):
    if isinstance(tensor, ndarray):
        result = tensor
        for i,e in enumerate(tensor):
            result[i] = equal_great(tensor[i])
        return result
    else:
        if tensor >= 0:
            return True
        else:
            return False
def equal_less(tensor):
    if isinstance(tensor, ndarray):
        result = tensor
        for i,e in enumerate(tensor):
            result[i] = equal_less(tensor[i])
        return result
    else:
        if tensor <= 0:
            return True
        else:
            return False

# Server setup functions
def set_database(env_variable):
    url = urlparse(environ[env_variable])
    conn = connect(
                dbname=url.path[1:],
                user=url.username,
                password=url.password,
                host=url.hostname,
                port=url.port
                )
    return conn
def schedule_buyer_confirmation():
    scheduler = BackgroundScheduler()
    scheduler.add_job(func=check_buyer_confirmation, trigger="interval", hours=3)
    scheduler.start()
    register(lambda: scheduler.shutdown())
    return scheduler
def check_result(tensor, dimensions, condition):
    tensor_size = tensor.shape
    flag = True
    for i, dim in enumerate(tensor_size):
        if dim > dimensions[i]/condition:
            flag = False
    return flag
def check_buyer_confirmation():
    global my_address
    global pending_confirmation
    print("Checking buyer receipt confirmation")
    for buyer in reversed(pending_confirmation):
        flag_pay = contract_deployed.functions.f_buyer_payed(buyer).call()
        if not flag_pay:
            flag_sent_decrypted = contract_deployed.functions.f_provider_sent_result(buyer).call()
            if flag_sent_decrypted:
                flag_received = contract_deployed.functions.f_buyer_received_result(buyer).call()
                if not flag_received:
                    key = contract_deployed.functions.reclaim_value(buyer).buildTransaction({
                        'nonce': web3.eth.getTransactionCount(my_address),
                        'gas': 1728712,
                        'gasPrice': web3.toWei('21','gwei')
                    })
                    signed_tx = web3.eth.account.signTransaction(key, private_key=private_key)
                    hash_tx = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
                    receipt_tx = web3.eth.waitForTransactionReceipt(hash_tx)
                    pending_confirmation.remove(buyer)



from requests import post
from requests import get
from json import loads

from hesystem.essentials import retrieve_user_info
from hesystem.essentials import PaillierTensor
from hesystem.essentials import EncryptedNumber
from hesystem.essentials import MultiDimensionalArrayEncoder
from hesystem.essentials import hinted_tuple_hook
from hesystem.essentials import serialize_paillier
from hesystem.essentials import deserialize_paillier
from hesystem.essentials import deserialize
from hesystem.essentials import to_paillier
from hesystem.essentials import set_web3
from hesystem.essentials import torch
from hesystem.essentials import random

# User information container
class User_Data():
    def __init__(self, address, private_key, endpoint_url):
        self.address = address
        self.private_key = private_key
        self.endpoint_url = endpoint_url
        self.web3 = set_web3(endpoint_url, address)
        print('User information ready!')
        print('===========================')

# Client-Server interaction functions
def retrieve_contract_web(web3, url):
    response = get(url)
    data = response.json()
    contract_address = data['address']
    contract_value = data['value']
    contract_abi = data['abi']
    address = web3.toChecksumAddress(contract_address)
    contract = web3.eth.contract(address=address,abi=contract_abi)
    return contract, contract_value
def buy_data(User, value):
    contract = User.contract
    web3 = User.web3
    key = contract.functions.buy_data().buildTransaction({
        'nonce': web3.eth.getTransactionCount(User.address),
        'value': value,
        'gas': 1728712,
        'gasPrice': web3.toWei('21', 'gwei')
    })
    signed_tx = web3.eth.account.signTransaction(key, private_key=User.private_key)
    hash_tx = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
    receipt_tx = web3.eth.waitForTransactionReceipt(hash_tx)
    if receipt_tx.status:
        url = contract.functions.data_link().call()
        return retrieve_data(User.address, url)
    else:
        buy_confirmation = contract.functions.f_buyer_payed(User.address).call()
        if buy_confirmation:
            url = contract.functions.data_link().call()
            return retrieve_data(User.address, url)
        else:
            raise Exception('The buy_data function was not executed correctly')

# Client visible functions
def initialize():
    print('Setting up user information')
    print('===========================')
    load = input('Do you want to load info from file? (Y/N) ')
    if load == 'Y' or load == 'y':
        user_path = input('Enter your information file path: ')
        address, private_key, endpoint_url = retrieve_user_info(user_path)
        return User_Data(address, private_key, endpoint_url)
    else:
        address = input('Enter your address (0x514AEa42dA89B7856C81bdAA4A20BD7D64EbA8E4): ')
        private_key = input('Enter your private_key (5D232502101181CADEF51F19294A981E22D2DCA38AB031E9BA6EE12F512263BA): ')
        endpoint_url = input('Enter your endpoint_url (https://ropsten.infura.io/v3/52c1338c6a2220d4a18dfef32ba53c2a): ')
        save = input('Do you want to save this info in a file? (Y/N) ')
        if save == 'Y' or save == 'y':
            user_path = input('Enter your information file path: ')
            user_file = open(user_path, 'w')
            user_file.write('address:'+address+'\n')
            user_file.write('private_key:'+private_key+'\n')
            user_file.write('endpoint_url:'+endpoint_url+'\n')
            user_file.close()
        return User_Data(address, private_key, endpoint_url)
def request_data(User):
    details_url = input('Enter the data details url: ')
    contract, value = retrieve_contract_web(User.web3, details_url)
    User.contract = contract
    data, public_key = buy_data(User, value)
    print('Data has been received!')
    print('===========================')
    return data, public_key
def request_test(User):
    details_url = input('Enter the data details url: ')
    contract, value = retrieve_contract_web(User.web3, details_url)
    test_url = contract.functions.test_link().call()
    User.contract = contract
    data, public_key = retrieve_data(User.address, test_url)
    print('Test data has been received!')
    print('===========================')
    return data, public_key
def request_result(User, result, public_key):
    print('Real result has been requested!')
    print('===========================')
    contract = User.contract
    web3 = User.web3
    key = contract.functions.send_result().buildTransaction({
        'nonce': web3.eth.getTransactionCount(User.address),
        'gas': 1728712,
        'gasPrice': web3.toWei('21','gwei')
    })
    signed_tx = web3.eth.account.signTransaction(key, private_key=User.private_key)
    hash_tx = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
    receipt_tx = web3.eth.waitForTransactionReceipt(hash_tx)
    if receipt_tx.status:
        decrypted_result = retrieve_result(User, result, public_key)
        if isinstance(decrypted_result, torch.Tensor):
            key = contract.functions.confirm_result().buildTransaction({
                'nonce': web3.eth.getTransactionCount(User.address),
                'gas': 1728712,
                'gasPrice': web3.toWei('21','gwei')
            })
            signed_tx = web3.eth.account.signTransaction(key, private_key=User.private_key)
            hash_tx = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
            receipt_tx = web3.eth.waitForTransactionReceipt(hash_tx)
        else:
            print('The result is not in Torch format. The transaction has been reversed.')
            # Ligitation
            key = contract.functions.start_ligitation_buyer().buildTransaction({
                'nonce': web3.eth.getTransactionCount(User.address),
                'gas': 1728712,
                'gasPrice': web3.toWei('21','gwei')
            })
            signed_tx = web3.eth.account.signTransaction(key, private_key=User.private_key)
            hash_tx = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
            receipt_tx = web3.eth.waitForTransactionReceipt(hash_tx)
        return decrypted_result
    else:
        raise Exception('The send_result blockchain function was not executed correctly. Did you already request and receive a result?')

    return

# Data handling useful functions
def retrieve_data(my_address, url):
    response = post(url, json={'address': my_address})
    json_obj = response.content
    data_obj = loads(json_obj, object_hook=hinted_tuple_hook)
    try:
        data = deserialize_paillier(data_obj)
        public_key = data.child.pubkey
        # Setting comparison functions to PaillierTensor
        set_gt(url, public_key, PaillierTensor)
        set_lt(url, public_key, PaillierTensor)
        set_eq(url, public_key, PaillierTensor)
        set_ge(url, public_key, PaillierTensor)
        set_le(url, public_key, PaillierTensor)
        set_transpose(PaillierTensor)
        set_dim(PaillierTensor)
        # Setting comparison functions to EncryptedNumber
        set_gt(url, public_key, EncryptedNumber)
        set_lt(url, public_key, EncryptedNumber)
        set_eq(url, public_key, EncryptedNumber)
        set_ge(url, public_key, EncryptedNumber)
        set_le(url, public_key, EncryptedNumber)
        return data, public_key
    except:
        return data_obj, data_obj
def retrieve_result(User, result, public_key):
    random_data = random.randint(100)
    result = result + random_data
    result_obj = serialize_paillier(to_paillier(result, public_key))
    json_obj = MultiDimensionalArrayEncoder().encode(result_obj)
    data = {'buyer': User.address, 'obj': json_obj}
    result_url = User.contract.functions.result_link().call()
    response = post(result_url, json=data)
    string_result = response.content
    try:
        print('Result has been received!')
        print('===========================')
        decrypted_result = deserialize(string_result)
        decrypted_result = decrypted_result - random_data
    except:
        decrypted_result = loads(string_result)
    return decrypted_result
def retrieve_test_result(User, result, public_key):
    random_data = random.randint(100)
    result = result + random_data
    result_obj = serialize_paillier(to_paillier(result, public_key))
    json_obj = MultiDimensionalArrayEncoder().encode(result_obj)
    data = {'buyer': User.address, 'obj': json_obj}
    result_url = User.contract.functions.test_result_link().call()
    response = post(result_url, json=data)
    string_result = response.content
    try:
        print('Result has been received!')
        print('===========================')
        decrypted_result = deserialize(string_result)
        decrypted_result = decrypted_result - random_data
    except:
        decrypted_result = loads(string_result)
    return decrypted_result
def set_gt(url, public_key, ModifiedClass):
    def client_gt(self, *args, **kwargs):
        tensor = self - args[0]
        serialized_tensor = serialize_paillier(to_paillier(tensor, public_key))
        json_obj = MultiDimensionalArrayEncoder().encode(serialized_tensor)
        params = {'obj': json_obj, 'function': '>'}
        r = post(url=url+'/comparetensor', json= params)
        return deserialize(r.content)
    ModifiedClass.__gt__ = client_gt
def set_lt(url, public_key, ModifiedClass):
    def client_lt(self, *args, **kwargs):
        tensor = self - args[0]
        serialized_tensor = serialize_paillier(to_paillier(tensor, public_key))
        json_obj = MultiDimensionalArrayEncoder().encode(serialized_tensor)
        params = {'obj': json_obj, 'function': '<'}
        r = post(url=url+'/comparetensor', json= params)
        return deserialize(r.content)
    ModifiedClass.__lt__ = client_lt
def set_eq(url, public_key, ModifiedClass):
    def client_eq(self, *args, **kwargs):
        tensor = self - args[0]
        serialized_tensor = serialize_paillier(to_paillier(tensor, public_key))
        json_obj = MultiDimensionalArrayEncoder().encode(serialized_tensor)
        params = {'obj': json_obj, 'function': '='}
        r = post(url=url+'/comparetensor', json= params)
        return deserialize(r.content)
    ModifiedClass.__eq__ = client_eq
def set_ge(url, public_key, ModifiedClass):
    def client_ge(self, *args, **kwargs):
        tensor = self - args[0]
        serialized_tensor = serialize_paillier(to_paillier(tensor, public_key))
        json_obj = MultiDimensionalArrayEncoder().encode(serialized_tensor)
        params = {'obj': json_obj, 'function': '>='}
        r = post(url=url+'/comparetensor', json= params)
        return deserialize(r.content)
    ModifiedClass.__ge__ = client_ge
def set_le(url, public_key, ModifiedClass):
    def client_le(self, *args, **kwargs):
        tensor = self - args[0]
        serialized_tensor = serialize_paillier(to_paillier(tensor, public_key))
        json_obj = MultiDimensionalArrayEncoder().encode(serialized_tensor)
        params = {'obj': json_obj, 'function': '<='}
        r = post(url=url+'/comparetensor', json= params)
        return deserialize(r.content)
    ModifiedClass.__le__ = client_le
def set_transpose(ModifiedClass):
    def functranspose(self):
        return self.transpose()
    ModifiedClass.t = functranspose
def set_dim(ModifiedClass):
    def funcdim(self):
        return len(self.shape)
    ModifiedClass.dim = funcdim



from hesystem.client import initialize
# Purchase functions
from hesystem.client import request_data
from hesystem.client import request_result
# Testing functions
from hesystem.client import request_test
from hesystem.client import retrieve_test_result
